;--------------------------------------------------------------------
; Conway's Game of Life simulator.
; (C) Copyright by Marek Karcz 2013.
; All rights reserved.
;
; History of revisions:
; 
; * Some summer day in 2013
;     Initial revision.
; * 9/13/2013
;     Random colony generator in start menu.
; * 9/14/2013
;     Framed pop-up prompt messages.
; * 9/15/2013
;     Ability to create random colony directly in editor.
;
; TO DO:
; 1) Add statistics report (iterations, died, born, population).
; 2) Add stalled colony detection (same population for some
;    predefined amount of iterations should trigger prompt to
;    continue or stop simulation).
; 3) Add colony file name lookup (directory).
;
; BUGS:
;   None known at the moment.
; 
;--------------------------------------------------------------------
!to "life.prg",cbm

BLACK = 0
GREEN = 5
LCELL = $51       ; 'alive' cell screen code
ECELL = 32        ; empty cell screen code
BCELL = 96        ; 'born' cell screen code
DCELL = $57       ; 'died' cell screen code
MINCOL = 0        ; minimum column #
MAXCOL = 39       ; maximum column #
MINROW = 0        ; minimum row #
MAXROW = 24       ; maximum row #
TDELAY = 2 * 4    ; time delay (quarters of a second)

!source "bincodes.asm"

; BASIC startup
*=$0800
!by $00,$0c,$08,$0a,$00,$9e,$32,$30,$36,$34,$00,$00,$00,$00

; application code
*=$0810

    jmp main

!source "system.asm"

!zone MainZone

main
    jsr SystemZone.clrscr
    jsr setcolors
    jsr SystemZone.upcaseon
    jsr SystemZone.disshcbm
    jsr prnwelcome

    lda #TDELAY
    jsr SystemZone.delay
    
    clc
    ldx #22
    ldy #0
    jsr PLOT
    jsr presskeytocont
    
    lda #0
    jsr pophelp1

main00    
    jsr init
    beq prgend00
    jsr edit
    
    jsr w4kpress    
loop
    jsr next
    jsr disp
    jsr SystemZone.getkey
    cmp #'P'
    bne loop0
    jsr w4kpress
loop0    
    cmp #'E'
    bne loop2
    ; go2edit
    jsr edit
    clc
    bcc loop
loop2    
    cmp #'Q'
    beq prgend
    cmp #'M'
    beq gtmenu1
    cmp #'H'
    bne loop
    ; pop-up help screen
    lda #1        ; set Acc!=0 to store screen
    jsr pophelp1
    clc
    bcc loop
gtmenu1    ; go to start menu1
    jsr popconfirm
    bne loop
    clc
    bcc main00
prgend 
    jsr popconfirm
    bne loop
prgend00    
    jsr cleanup
    rts
    
; pop press any key to cont.
; wait for a keypress
presskeytocont
    lda #<anykey
    sta ZPGA1
    lda #>anykey
    sta ZPGA1+1
    jsr SystemZone.prnstrzpga1
    jsr w4kpress
    rts
    
; pop-up help screen
; if Acc != 0, store/restore screen
pophelp1
    tax
    ; store ZPGA1 to stack
    lda ZPGA1
    pha
    lda ZPGA1+1
    pha
    txa
    pha           ; store Acc (storescr flag) to stack
    beq ph00    
    jsr storescr
ph00
    lda #<help1
    sta ZPGA1
    lda #>help1
    sta ZPGA1+1
    jsr SystemZone.prnstrzpga1
    jsr presskeytocont  
    pla           ; restore Acc (sorescr flag) from stack
    beq ph01
    jsr restorescr
ph01
    ; restore ZPGA1 from stack
    pla
    sta ZPGA1+1
    pla
    sta ZPGA1
    rts

; store screen in buffer
storescr    
    lda #<MSCR      ; save the current screen in buffer
    sta ZPGA1
    lda #>MSCR
    sta ZPGA1+1
    lda #<colb
    sta ZPGA2
    lda #>colb
    sta ZPGA2+1
    ldx #<1000
    ldy #>1000
    jsr SystemZone.memcpy
    rts
    
; restore screen from buffer
restorescr
    lda #<MSCR
    sta ZPGA2
    lda #>MSCR
    sta ZPGA2+1
    lda #<colb
    sta ZPGA1
    lda #>colb
    sta ZPGA1+1
    ldx #<1000
    ldy #>1000
    jsr SystemZone.memcpy
    rts
    
; save screen in buffer and pop confirmation prompt
; if answered Y, return 0 in Acc
; if answered N, restore screen from buffer, return 1 in Acc
popconfirm
    jsr storescr
    lda #9
    jsr areyousure  ; and pop-up a confirmation prompt
    bne popcnotsure
    rts
popcnotsure
    jsr restorescr
    lda #1
    rts

; print welcome banner
prnwelcome
    lda #<scrwelcome
    sta ZPGA2
    lda #>scrwelcome
    sta ZPGA2+1
    lda #5
    jsr popscr
    rts
        
; wait for keypress
; store key code in keyp buffer and return it in Acc
w4kpress
    jsr SystemZone.getkey
    beq w4kpress
    sta keyp
    lda keyp
    rts

; initialization
; Return 0 in Acc, if Q - quit option is selected
init
    jsr setcolors
    jsr clrdisp
    jsr initcolb
    jsr demo
    rts
    
; initialize colb buffer
initcolb
    lda #<colb
    sta ZPGA1
    lda #>colb
    sta ZPGA1+1
    lda #0
    ldx #<1000
    ldy #>1000
    jsr SystemZone.fillmem
    rts
    
; colors
setcolors
    lda #BLACK
    sta BCOL
    sta BGCOL
    lda #GREEN
    sta CCOL
    jsr SystemZone.fillmcol
    rts
    
; fill screen memory with ECELL codes
clrdisp

    lda #ECELL
    jsr SystemZone.fillscr
    rts

; colony editor
edit
    jsr filleditscr ; fill editor's screen
    lda #0          ; initialize row and column counters
    sta xind
    sta yind
    lda #<MSCR
    sta ZPGA1
    lda #>MSCR
    sta ZPGA1+1
    jsr setrevcursor
edloop
    jsr w4kpress
    cmp #CRSUP
    bne edchk1
    ; cursor-up pressed, check if in 1st row
    lda xind
    beq edchk1      ; cursor in 1st row, ignore CRSUP pressed
    ; cursor-up handler
    jsr unsetrevcursor
    jsr sb40zpga1   ; update cursor position pointer
    dec xind        ; decrement row counter
    jsr setrevcursor
    clc
    bcc edloop    
edchk1
    lda keyp
    cmp #CRSDOWN
    bne edchk2
    ; cursor-down pressed, check if in last row
    lda xind
    cmp #MAXROW
    beq edchk2      ; cursor in last row, ignore CRSDOWN pressed
    ; cursor-down handler
    jsr unsetrevcursor
    jsr add40zpga1  ; update cursor position pointer
    inc xind        ; increment row counter
    jsr setrevcursor
    clc
    bcc edloop    
edchk2
    lda keyp
    cmp #CRSLEFT
    bne edchk3
    ; cursor-left pressed, check if in 1st column
    lda yind
    beq edchk3      ; cursor in 1st column, ignore CRSLEFT pressed
    ; cursor-left handler
    jsr unsetrevcursor
    jsr SystemZone.deczpga1    ; update cursor position pointer
    dec yind        ; decrement column counter
    jsr setrevcursor
    clc
    bcc edloop    
edrestart
    clc
    bcc edit    
edchk3
    lda keyp
    cmp #CRSRIGHT
    bne edchk4
    ; cursor-right pressed, check if in the last column
    lda yind
    cmp #MAXCOL
    beq edchk4      ; last column, ignore CRSRIGHT pressed
    ; cursor-right handler
    jsr unsetrevcursor
    jsr SystemZone.inczpga1    ; update cursor position pointer
    inc yind        ; increment column counter
edsetcrs    
    jsr setrevcursor
    clc
    bcc edloop
edchk4
    lda keyp
    cmp #$20        ; space is a set/unset key in editor
    bne edchk6
    ; set/unset handler
    ldx #0
    lda (ZPGA1,x)
    cmp #('-'+128)
    bne edchk5
    lda #'*'
    sta (ZPGA1,x)
    clc
    bcc edsetcrs
edchk5    
    cmp #('*'+128)
    bne edchk6
    lda #'-'
    sta (ZPGA1,x)
    clc
    bcc edsetcrs
edchk6
    lda keyp
    cmp #'Q'
    beq eddone
    cmp #'H'
    bne edchk7
    ; pop-up help screen
    lda #1        ; set Acc!=0 to trigger storescr
    jsr pophelp1
    clc
    bcc edsetcrs
edchk7    
    ;lda keyp
    cmp #'C'
    beq edclr
    cmp #'R'
    bne edsetcrs
    jsr popconfirm
    bne edsetcrs
    jsr randcol
    clc
    bcc edrestart
edclr
    jsr unsetrevcursor
    jsr clreditor
    clc
    bcc edrestart
eddone    ; edit done
    ; save?
    jsr unsetrevcursor
    lda ZPGA1
    pha
    lda ZPGA1+1
    pha
    jsr savecolony
    pla
    sta ZPGA1+1
    pla
    sta ZPGA1
    ; convert to simulation format
    ;jsr unsetrevcursor
    jsr conved2sim
    rts
    
; load colony from file
loadcolony
    lda #<fnameprompt2 ; ask for file name
    sta ZPGA1
    lda #>fnameprompt2
    sta ZPGA1+1
    jsr SystemZone.prnstrzpga1
    jsr inputfname
    sta fnlen         ; store file name length
    jsr loadcolb
    jsr decodecolb2ed ; decode binary format to editor's format
    rts
    
; decode colony in binary format in colb to screen format
decodecolb2ed
    lda #<MSCR
    sta ZPGA1
    lda #>MSCR
    sta ZPGA1+1
    lda #125
    sta gpbyt     ; encoded bytes counter
    lda #8
    sta gpct+1    ; bit counter
    ldx #0
    lda colb,x    ; load 1-st encoded byte
    sta gpct      ; store encoded byte
    ldy #0
    lda gpbyt
    bne dccolbl1
    rts
dccolbl1
    asl gpct      ; shift left, bit 7 to Carry
    bcc dccolbl2  ; Carry not set, go to dccolbl2
    lda #LCELL    ; load life cell code
    bne dccolbl3  ; jump to dccolbl3
dccolbl2
    lda #ECELL    ; load empty cell code
dccolbl3    
    sta (ZPGA1),y ; put in screen memory
    iny           ; increment ZPG index
    dec gpct+1    ; decrement bit counter
    bne dccolbl4  ; not zero, still the same byte
    dec gpbyt     ; new byte, decrement bytes counter
    lda #8
    sta gpct+1    ; reload bit counter
    ldy #0        ; reload ZPG index
    jsr add8zpga1 ; add 8 to ZPGA1 pointer (next 8 cells)
    inx           ; increment encoded bytes pointer
    lda colb,x    ; load encoded byte
    sta gpct      ; store encoded byte    
dccolbl4
    lda gpbyt
    bne dccolbl1
    rts
    
; confirm with user and save the colony    
savecolony
    jsr storescr    ; store screen to buffer (colb)
    lda #9
    jsr asktosave   ; prompt to save
    bne scend       ; no saving
    ; colony saving routine
    lda #<scrfnameprompt
    sta ZPGA2
    lda #>scrfnameprompt
    sta ZPGA2+1
    lda #9
    jsr popscr
    clc
    ldx #10
    ldy #21
    jsr PLOT
    jsr inputfname
    sta fnlen         ; store file name length
    jsr restorescr    ; restore screen from buffer (colb)
    jsr encodeed2colb ; convert editor's format to binary format in colb location
    jsr savecolb      ; save file
    rts
scend    
    jsr restorescr    ; restore screen from buffer (colb)
    rts
    
; input file name, up to 12-char long
; return the length in Acc
inputfname
    jsr prncursor ; print cursor to the screen
    ldx #0
    stx gpbyt
ifnl01
    jsr SystemZone.getkey
    beq ifnl01
    cmp #NL       ; is it New Line code?
    beq ifnend    ; yes, wrap it up
    cmp #DELETE   ; is it DEL (Backspace) key?
    beq ifn001    ; yes, do the backspace action
    jsr isalnum   ; check if aphpa-numeric
    beq ifnl01    ; no, get another key press
    bne ifnl02    ; yes, add the character
ifn001    
    lda gpbyt
    beq ifnl01    ; do nothing, there are no characters yet
    dec gpbyt     ; decrement file name buffer index
    lda #CRSLEFT  ; delete preceeding character
    jsr CHROUT
    jsr CHROUT
    lda #' '
    jsr CHROUT
    jsr CHROUT
    lda #CRSLEFT
    jsr CHROUT
    jsr CHROUT
    jsr prncursor ; output cursor
    clc
    bcc ifnl01    ; branch to the beginning of the loop
ifnl02    
    ldx gpbyt     ; if at the buffer limit
    cpx #12       ; accept only RETURN or DELETE
    beq ifnl01    ; and ignore the alpha-numeric character
    pha           ; store character code on stack
    lda #CRSLEFT  ; move left
    jsr CHROUT
    pla           ; restore character code from stack
    jsr CHROUT    ; print character
    ldx gpbyt
    sta fnamb,x   ; store character code in file name buffer
    inx           ; increment file name buffer index
    stx gpbyt     ; store file name buffer index
    jsr prncursor ; print cursor
    clc
    bcc ifnl01    ; branch to the beginning of the loop
ifnend
    lda gpbyt     ; check if there are characters entered
    bne ifnend00  ; there are characters
    rts           ; there are no characters
ifnend00    
    ; add extension .COL
    tax
    ldy #0
ifnaddextl00
    lda fnamext,y
    beq ifnaddextdone
    sta fnamb,x
    inx
    iny
    bne ifnaddextl00
ifnaddextdone
    inx
    lda #0        ; terminate with NULL
    sta fnamb,x
    dex
    txa           ; return length in Acc
    rts
    
; print cursor character
prncursor
    lda #181
    jsr CHROUT
    rts
    
; check if character code in Acc is alphanumeric (32-127)
; return 0 in Acc if not, character code if yes
isalnum
    pha             ; save char on stack
    sec
    cmp #32         ; subtract 32 (Acc < 32 ?)
    bcs ian01       ; no, check if less than 128
ian00    
    pla             ; clean stack
    lda #0          ; not alpha-numeric, return 0
    rts
ian01
    pla             ; re-load char and re-push to stack
    pha
    sec
    cmp #128        ; subtract 128 (Acc < 128 ?)
    bcs ian00       ; no, not alpha-numeric
    pla             ; restore character from stack to Acc
    rts

COLBEND = colb+126

; save binary-encoded colony in colb to file
savecolb
    lda #1
    ldx #8
    ldy #0
    jsr SETLFS
    lda fnlen
    ldx #<fnamb
    ldy #>fnamb
    jsr SETNAM
    lda #<colb
    sta ZPGA1
    lda #>colb
    sta ZPGA1+1
    ldx #<COLBEND
    ldy #>COLBEND
    lda #ZPGA1
    jsr SAVE
    ; TO DO: add error handling
    rts
    
; load binary-encoded colony from file to colb
loadcolb
    lda #1
    ldx #8
    ldy #0
    jsr SETLFS
    lda fnlen
    ldx #<fnamb
    ldy #>fnamb
    jsr SETNAM
    lda #0
    ldx #<colb
    ldy #>colb
    jsr LOAD
    ; TO DO: add error handling
    rts
    
; mask bits array for colony encoding routine    
encodearr

    !byte B00000000
    !byte B00000001
    !byte B00000010
    !byte B00000100
    !byte B00001000
    !byte B00010000
    !byte B00100000
    !byte B01000000
    !byte B10000000
    
; convert screen (editor's format) into binary format
; store in colb buffer
encodeed2colb
    lda #<MSCR
    sta ZPGA1
    lda #>MSCR
    sta ZPGA1+1
    lda #0
    sta gpct        ; gpct - encoded byte
    sta gpct+1      ; index to colb buffer (encoded colony)
    lda #8
    sta gpbyt       ; gpbyt - bit counter
ee2cloop000
    ldy #0
    lda (ZPGA1),y   ; load byte from screen memory
    cmp #'*'        ; is it live cell?
    bne ee2cloop002 ; no, branch to ee2cloop002
    lda gpct        ; load encoded byte to Acc
    ldx gpbyt       ; load bit counter to X
    ora encodearr,x ; set the bit in Acc masked by encodearr,X
    sta gpct        ; store updated encoded byte
ee2cloop002    
    dec gpbyt       ; decrement bit counter
    lda gpbyt
    beq ee2cloop001 ; if byte counter is 0, branch to ee2cloop001
ee2cloop000a    
    jsr SystemZone.inczpga1 ; increment ZPG1 (screen memory)
    lda ZPGA1       ; check if end of screen memory
    cmp #<MESCR
    bne ee2cloop000
    lda ZPGA1+1
    cmp #>MESCR
    beq ee2cfinished
    clc
    bcc ee2cloop000 ; do next bit
ee2cloop001         ; single byte encoding finished
    ldx gpct+1      ; load encoded colony buffer index
    lda gpct        ; load encoded byte
    sta colb,x      ; store byte in the encoded colony buffer
    inc gpct+1      ; increment index to encoded colony buffer
    lda #8          ; reload bit counter
    sta gpbyt
    lda #0          ; reset encoded byte
    sta gpct
    clc
    bcc ee2cloop000a  ; do the next bit
ee2cfinished
    rts
    
; fill editor's screen with data in proper format
; or in other words, convert simulator's screen codes
; to editor's screen codes
filleditscr
    lda #<MSCR
    sta ZPGA1
    lda #>MSCR
    sta ZPGA1+1
filledl1
    ldx #0
    lda (ZPGA1,x)
    cmp #ECELL
    bne filledl2    ; not an empty-cell, jump to filledl2
    lda #'-'        ; load editor's equivalent of empty cell code
    sta (ZPGA1,x)
    clc
    bcc filledl3
filledl2
    lda (ZPGA1,x)
    cmp #LCELL
    bne filledl3    ; not a live cell, jump to filledl3
    lda #'*'        ; load editor's equivalent of live cell
    sta (ZPGA1,x)
filledl3
    lda ZPGA1
    cmp #<MESCR
    bne filledl4
    lda ZPGA1+1
    cmp #>MESCR
    beq filledend
filledl4    
    jsr SystemZone.inczpga1
    clc
    bcc filledl1
filledend    
    rts
    
; clear colony in editor
clreditor
    jsr popconfirm
    bne clredno
    jsr clrdisp
clredno    
    rts
    
; convert editor's screen codes to simulator's format    
conved2sim
    lda #<MSCR
    sta ZPGA1
    lda #>MSCR
    sta ZPGA1+1
convedl1
    ldx #0
    lda (ZPGA1,x)
    cmp #'-'
    bne convedl2    ; not an empty-cell, jump to convedl2
    lda #ECELL      ; load simulator's equivalent of empty cell code
    sta (ZPGA1,x)
    clc
    bcc convedl3
convedl2
    lda (ZPGA1,x)
    cmp #'*'
    bne convedl3    ; not a live cell, jump to convedl3
    lda #LCELL      ; load simulator's equivalent of live cell
    sta (ZPGA1,x)
convedl3
    lda ZPGA1
    cmp #<MESCR
    bne convedl4
    lda ZPGA1+1
    cmp #>MESCR
    beq convedend
convedl4    
    jsr SystemZone.inczpga1
    clc
    bcc convedl1
convedend    
    rts
    
; colony editor helper routine    
; set reverse mode on a character currently under cursor's position
setrevcursor
    ldx #0
    lda (ZPGA1,x)
    ora #$80
    sta (ZPGA1,x)
    rts
    
; colony editor helper routine    
; unset reverse mode on a character currently under cursor's position
unsetrevcursor
    ldx #0
    lda (ZPGA1,x)
    and #$7f
    sta (ZPGA1,x)
    rts
    
; update the colony on the screen
; this routine should be called after 'jsr next'
; converts DCELL to ECELL
; converts BCELL to LCELL
disp
    lda #<MSCR
    sta ZPGA1
    lda #>MSCR
    sta ZPGA1+1
    lda #<(MSCR+256)
    sta ZPGA2
    lda #>(MSCR+256)
    sta ZPGA2+1
    ldy #0
    sty gpct
displ0
    lda (ZPGA1),y
    cmp #DCELL
    bne displ01
    lda #ECELL
    sta (ZPGA1),y
    clc
    bcc displ02
displ01
    cmp #BCELL
    bne displ02
    lda #LCELL
    sta (ZPGA1),y
displ02    
    lda (ZPGA2),y
    cmp #DCELL
    bne displ001
    lda #ECELL
    sta (ZPGA2),y
    clc
    bcc displ2
displ001
    cmp #BCELL
    bne displ2
    lda #LCELL
    sta (ZPGA2),y
displ2
    iny
    bne displ0
    inc gpct
    lda gpct
    cmp #2
    beq displend
    lda #<(MSCR+512)
    sta ZPGA1
    lda #>(MSCR+512)
    sta ZPGA1+1
    lda #<(MSCR+768)
    sta ZPGA2
    lda #>(MSCR+768)
    sta ZPGA2+1
    ldy #0    
    beq displ0
displend    
    rts
    
; calculate the next generation    
; Theory of operation:
; The screen memory is scanned from start to end.
; Each row is iterated with pointer ZPGA1 and index Y.
; Addressing mode (ZPGA1),Y is used. Register Y is
; reloaded with column counter yind, which is incremented
; accordingly at the end of the loop,
; then ZPGA1 is increased by 40 at the start of the next row.
; The number of neighbors is calculated for each cell.
; Depending on the # of neighbors, the cell is encoded
; to either die (DCELL), be born (BCELL) or remain unchanged.
; After this procedure, the 'disp' routine should be called
; to convert died cells to empty spaces and born cells (BCELL)
; to live cells (LCELL).
next
    ldx #0
    stx xind  ; initialize row index
    stx yind  ; initialize column index
    stx nbct  ; initialize neighbor cells counter
    lda #<MSCR  ; initialize ZPG address buffer
    sta ZPGA1   ; and current cell address counter
    sta smct
    lda #>MSCR
    sta ZPGA1+1
    sta smct+1
nl0 ; left neighbor check    
    ldy yind
    beq nl1         ; skip left neighbor check if column=0
    dey             ; decrement ZPGA1 pointer (left side neighbor)
    lda (ZPGA1),y   ; load cell's value
    cmp #LCELL      ; if not a live cell
    beq nl000
    cmp #DCELL      ; or not a 'died' cell (in this iteration)
    bne nl00        ; jump to nl00
nl000    
    inc nbct        ; increment neighbors count
nl00    
    ldy yind
nl1 ; right neighbor check
    cpy #MAXCOL     ; if cell at the far right side
    beq nl2         ; then skip right neighbor check
    iny             ; increment ZPGA1 pointer (right side neighbor)
    lda (ZPGA1),y   ; load cell's value
    cmp #LCELL      ; if not a live cell
    beq nl111
    cmp #DCELL      ; or not a 'died' cell (in this iteration)
    bne nl2        ; jump to nl2
nl111    
    inc nbct        ; increment neighbors count
nl2 ; north neighbors check
    lda xind
    beq nl4         ; skip north neighbors check if row=0
    jsr sb40zpga1   ; subtract 40 from ZPGA1 (north neighbor)
    ldy yind
    lda (ZPGA1),y   ; load cell's value
    cmp #LCELL      ; if not a live cell
    beq nl222
    cmp #DCELL      ; or not a 'died' cell (in this iteration)
    bne nl22        ; jump to nl22
nl222    
    inc nbct        ; increment neighbors count
nl22  ; north-west (NW) neighbor check
    ldy yind
    beq nl3         ; skip NW neighbor check if column=0
    dey             ; decrement ZPGA1 pointer (NW side neighbor)
    lda (ZPGA1),y   ; load cell's value
    cmp #LCELL      ; if not a live cell
    beq nl2220
    cmp #DCELL      ; or not a 'died' cell (in this iteration)
    bne nl220       ; jump to nl220
nl2220    
    inc nbct        ; increment neighbors count
nl220
    iny             ; return ZPGA1 pointer to north neighbor's cell
nl3
    cpy #MAXCOL     ; if cell at the far right side
    beq nl33        ; then skip the right neighbor check
    iny             ; increment ZPGA1 (right side neighbor)
    lda (ZPGA1),y   ; load cell's value
    cmp #LCELL      ; if not a live cell
    beq nl333
    cmp #DCELL      ; or not a 'died' cell (in this iteration)
    bne nl33        ; jump to nl33
nl333    
    inc nbct        ; increment neighbors count
nl33
    jsr rldsm2zpga1 ; reload smct to ZPGA1
nl4 ; south neighbor check
    lda xind
    cmp #MAXROW     ; if row at maximum,
    beq nl6         ; skip south neighbors check
    jsr add40zpga1  ; add 40 to ZPGA1 (now ZPGA1+Y -> south neighbor's cell)
    ldy yind
    lda (ZPGA1),y   ; load cell's value
    cmp #LCELL      ; if not a live cell
    beq nl444
    cmp #DCELL      ; or not a 'died' cell (in this iteration)
    bne nl44        ; jump to nl44
nl444    
    inc nbct        ; increment neighbors count
nl44  ; south-west (SW) neighbor check
    ldy yind
    beq nl5         ; skip SW neighbor check if column=0
    dey             ; decrement ZPGA1 (SW side neighbor)
    lda (ZPGA1),y   ; load cell's value
    cmp #LCELL      ; if not a live cell
    beq nl4440
    cmp #DCELL      ; or not a 'died' cell (in this iteration)
    bne nl440       ; jump to nl440
nl4440    
    inc nbct        ; increment neighbors count
nl440
    iny             ; return ZPGA1 pointer to south neighbor's cell
nl5
    cpy #MAXCOL     ; if cell at the far right side
    beq nl6         ; then skip the right neighbor check
    iny             ; increment ZPGA1 (SE side neighbor)
    lda (ZPGA1),y   ; load cell's value
    cmp #LCELL      ; if not a live cell
    beq nl55
    cmp #DCELL      ; or not a 'died' cell (in this iteration)
    bne nl6         ; jump to nl6
nl55    
    inc nbct        ; increment neighbors count    
nl6
    jsr rldsm2zpga1 ; reload smct to ZPGA1 (restore address of the row)
    lda yind        ; store old column# before incrementing
    sta yold        ; for further calculations
    inc yind        ; increment column#
    lda yind
    cmp #(MAXCOL+1)
    bne nl66
    lda #0
    sta yind
    jsr add40smct   ; increment smct (point to the next row of cells)
    inc xind        ; increment row#
nl66  ; now mark the cells to live/die/be born
    ldy yold
    lda (ZPGA1),y   ; load current cell
    cmp #LCELL      ; check if a live one
    beq nl666
    cmp #DCELL
    bne nl7         ; not a live one, jump to nl7
nl666 ; live cell, determine its fate
    lda nbct        ; load neighbors counter
    cmp #0          ; if nbct<2, cell dies
    beq nldie
    cmp #1
    beq nldie
    cmp #2
    beq nlnext
    cmp #3          ; if nbct==3, cell does not change
    beq nlnext      ; if nbct>3, cell dies
nldie
    lda #DCELL
    ldy yold
    sta (ZPGA1),y   ; update cell with 'died' code
    clc
    bcc nlnext      ; prepare for next cell
nl7 ; not a live cell, determine its fate
    lda nbct        ; load neighbors counter
    cmp #3          ; only if nbct==3, a new one is born
    bne nlnext
    lda #BCELL
    ldy yold
    sta (ZPGA1),y   ; update cell with 'born' status.
nlnext
    lda xind        ; check if the end
    cmp #(MAXROW+1)
    beq nlend       ; yes, jump to nlend
    jsr rldsm2zpga1 ; reload smct to ZPGA1
    lda #0          ; initialize neighbors counter
    sta nbct
    jmp nl0         ; do the next cell
nlend
    rts
    
; add 40 to smct
add40smct
    clc
    lda smct
    adc #40
    sta smct
    lda smct+1
    adc #0
    sta smct+1
    rts
    
; add 40 to ZPGA1    
add40zpga1
    clc
    lda ZPGA1
    adc #40
    sta ZPGA1
    lda ZPGA1+1
    adc #0
    sta ZPGA1+1
    rts
    
; subtract 40 from ZPGA1
sb40zpga1
    sec
    lda ZPGA1
    sbc #40
    sta ZPGA1
    lda ZPGA1+1
    sbc #0
    sta ZPGA1+1
    rts

; reload smct to ZPGA1
rldsm2zpga1
    lda smct
    sta ZPGA1
    lda smct+1
    sta ZPGA1+1
    rts
    
; finish and cleanup    
cleanup

    jsr SystemZone.enshftcbm
    jsr SystemZone.clrscr
    lda #<goodbye
    sta ZPGA1
    lda #>goodbye
    sta ZPGA1+1
    jsr SystemZone.prnstrzpga1
    rts
    
; demo
; select the built-in template for a new colony or load from file
; if option Q (quit program) is selected, return 0 in Acc.
demo
    lda #<menu1
    sta ZPGA1
    lda #>menu1
    sta ZPGA1+1
    jsr SystemZone.prnstrzpga1
demoloop    
    jsr w4kpress
    cmp #'1'
    bne demonot1
    jsr decodeshooter
    lda #1
    rts
demonot1
    cmp #'2'
    bne demonot2
    jsr decodeosc144
    lda #1
    rts
demonot2
    cmp #'3'
    bne demonot3
    jsr clrdisp
    lda #1
    rts
demonot3
    cmp #'4'
    bne demonot4
    jsr randcol
    lda #1
    rts
demonot4    
    cmp #'L'
    beq demoloadfile
    cmp #'Q'
    bne demoloop
    ; quit program
    jsr popconfirm
    beq demoquitsel
    jsr initcolb
    clc
    bcc demoloop
demoquitsel    
    lda #0
    rts
demoloadfile    
    jsr loadcolony
    lda #1
    rts

; generate random colony to colb buffer
; and decode it to the screen editor
randcol
    ; initialize random seed
    jsr RDTIM
    sta SystemZone.SEED0
    sty SystemZone.SEED1
    stx SystemZone.SEED2
    inx
    stx SystemZone.SEED3
    ; initialize memory counters
    lda #0
    sta gpbyt
randcol00     ; fill colb with binary encoded random colony
    jsr SystemZone.RAND
    ldx gpbyt
    lda SystemZone.SEED0
    sta colb,x
    inx
    lda SystemZone.SEED1
    sta colb,x
    inx
    lda SystemZone.SEED2
    sta colb,x
    inx
    lda SystemZone.SEED3
    sta colb,x
    inx
    stx gpbyt
    lda gpbyt
    cmp #128
    bne randcol00
    jsr decodecolb2ed
    rts    
    
; display "sorry, option not implemented" message
; wait few seconds, then return
;prnnotimpl
;    lda #<sorry1
;    sta ZPGA1
;    lda #>sorry1
;    sta ZPGA1+1
;    jsr SystemZone.prnstrzpga1
;    lda #TDELAY
;    jsr SystemZone.delay
;    rts
    
; pop confirmation prompt (are you sure (y/n))
; at line# specified in Acc.
; return 0 in Acc if confirmed (Y)
areyousure
    sta gpbyt
    lda #<scrareyousure
    sta ZPGA2
    lda #>scrareyousure
    sta ZPGA2+1
    lda gpbyt
    jsr popscrprompt
    rts
    
; pop a Y/N confirmation prompt
; at line# specified in Acc.
; prompt screen codes address: ZPG2
; return 0 in Acc if confirmed (Y)
popscrprompt
    jsr popscr
    jsr w4kpress
    cmp #'Y'
    beq popscrprompt00
    rts    
popscrprompt00
    lda #0
    rts    
    
; pop screen codes sequence pointed by ZPGA2 
; at line specified in Acc.
popscr
    sta gpbyt
    lda #<MSCR
    sta ZPGA1
    lda #>MSCR
    sta ZPGA1+1
popscr00
    lda gpbyt
    beq popscr01
    jsr add40zpga1
    dec gpbyt
    clc
    bcc popscr00
popscr01
    ldy #0
popscr02
    lda gpbyt
    lda (ZPGA2),y
    beq popscr03
    sta (ZPGA1),y
    iny
    bne popscr02
popscr03
    rts
    
; ask the confirm question (save colony (y/n))
; at line# specified in Acc.
; return 0 in Acc if confirmed (Y)
asktosave
    sta gpbyt
    lda #<scrsavecolony
    sta ZPGA2
    lda #>scrsavecolony
    sta ZPGA2+1
    lda gpbyt
    jsr popscrprompt
    rts

DCSHOOTENCBUF=dcshootenc+1

; decode and load shooter demo
decodeshooter
    lda #<DCSHOOTENCBUF
    sta ZPGA1
    lda #>DCSHOOTENCBUF
    sta ZPGA1+1
    lda #<colb
    sta ZPGA2
    lda #>colb
    sta ZPGA2+1
    ldx dcshootenc
    ldy #0
    jsr SystemZone.memcpy
    jsr decodecolb2ed
    rts
    
; add 8 to ZPGA1 pointer
add8zpga1
    clc
    lda ZPGA1
    adc #8
    sta ZPGA1
    lda ZPGA1+1
    adc #0
    sta ZPGA1+1
    rts
    
DCOSCP144ENCBUF = dcoscp144enc+1

; decode and load osc144 demo
decodeosc144
    lda #<DCOSCP144ENCBUF
    sta ZPGA1
    lda #>DCOSCP144ENCBUF
    sta ZPGA1+1
    lda #<colb
    sta ZPGA2
    lda #>colb
    sta ZPGA2+1
    ldx dcoscp144enc
    ldy #0
    jsr SystemZone.memcpy
    jsr decodecolb2ed
    rts
    
; demo colonies
    
; shooter colony, binary encoded    
dcshootenc    

    !byte 100    ; first byte is the number of encoded bytes that follow
    ; encoded colony
    !text B00000000,B00000000,B00000000,B00000000,B00000000
    !text B00000000,B00000000,B00000000,B00000000,B00000000
    !text B00000000,B00000000,B00000000,B00000000,B00000000
    !text B00000000,B00000000,B00000000,B00000000,B00000000
    !text B00000000,B00000000,B00000000,B00000000,B00000000
    !text B00000000,B00000000,B00000000,B11000000,B00011000
    !text B00000000,B00000000,B00000001,B01000000,B00011000
    !text B01100000,B00110000,B00000001,B10000000,B00000000
    !text B01100000,B01010000,B00000000,B00000000,B00000000
    !text B00000000,B01100000,B01100000,B00000000,B00000000
    !text B00000000,B00000000,B01010000,B00000000,B00000000
    !text B00000000,B00000000,B01000000,B00000000,B00000000
    !text B00000000,B00000000,B00000000,B00000000,B00001100
    !text B00000000,B00000000,B00000000,B00000000,B00001010
    !text B00000000,B00000000,B00000000,B00000000,B00001000
    !text B00000000,B00000000,B00000000,B00000000,B00000000
    !text B00000000,B00000000,B00000000,B00000000,B00000000
    !text B00000000,B00000000,B00000000,B01110000,B00000000
    !text B00000000,B00000000,B00000000,B01000000,B00000000
    !text B00000000,B00000000,B00000000,B00100000,B00000000

; oscillator 144 colony, binary encoded
dcoscp144enc

    !byte 95
    !text B00000011,B00000000,B00000000,B00000000,B11000000
    !text B00000011,B00000000,B00000000,B00000000,B11000000
    !text B00000000,B00000000,B00000000,B11000000,B00000000
    !text B00000000,B00000000,B00000001,B00100000,B00000000
    !text B00000000,B00000000,B00000000,B11000000,B00000000
    !text B00000000,B00000000,B00000000,B00000000,B00000000    
    !text B00000000,B00000000,B00011100,B00000000,B00000000
    !text B00000000,B00000000,B00010100,B00000000,B00000000
    !text B00000000,B00000000,B00011100,B00000000,B00000000    
    !text B00000000,B00000000,B00011000,B00000000,B00000000
    !text B00000000,B00000000,B00111000,B00000000,B00000000
    !text B00000000,B00000000,B00101000,B00000000,B00000000
    !text B00000000,B00000000,B00111000,B00000000,B00000000    
    !text B00000000,B00000000,B00000000,B00000000,B00000000
    !text B00000000,B00000011,B00000000,B00000000,B00000000
    !text B00000000,B00000100,B10000000,B00000000,B00000000
    !text B00000000,B00000011,B00000000,B00000000,B00000000    
    !text B00000011,B00000000,B00000000,B00000000,B11000000
    !text B00000011,B00000000,B00000000,B00000000,B11000000    
    
    
; application strings/texts

; framed welcome banner text, screen codes
scrwelcome
    !byte $20,$20,$20,$20,$20,$20,$20,$20,$70,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$6e,$20,$20,$20,$20,$20,$20,$20,$20,$20
    !byte $20,$20,$20,$20,$20,$20,$20,$20,$42,$03,$0f,$0e,$17,$01,$19,$27,$13,$20,$07,$01,$0d,$05,$20,$0f,$06,$20,$0c,$09,$06,$05,$42,$20,$20,$20,$20,$20,$20,$20,$20,$20
    !byte $20,$20,$20,$20,$20,$20,$20,$20,$42,$20,$13,$09,$0d,$15,$0c,$01,$14,$0f,$12,$2c,$20,$16,$05,$12,$3a,$20,$31,$2e,$30,$20,$42,$20,$20,$20,$20,$20,$20,$20,$20,$20
    !byte $20,$20,$20,$20,$20,$20,$20,$20,$42,$28,$03,$29,$20,$0d,$01,$12,$05,$0b,$20,$0b,$01,$12,$03,$1a,$20,$20,$32,$30,$31,$33,$42,$20,$20,$20,$20,$20,$20,$20,$20,$20
    !byte $20,$20,$20,$20,$20,$20,$20,$20,$42,$20,$01,$0c,$0c,$20,$12,$09,$07,$08,$14,$13,$20,$12,$05,$13,$05,$12,$16,$05,$04,$20,$42,$20,$20,$20,$20,$20,$20,$20,$20,$20
    !byte $20,$20,$20,$20,$20,$20,$20,$20,$6d,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$7d,$20,$20,$20,$20,$20,$20,$20,$20,$20
    !byte 0

; prompt text to press a key, cbm ascii
anykey

    !text NL
    !text "   >>> PRESS ANY KEY TO CONTINUE <<<",NL,0
    
; help screen text, cbm ascii
help1

    !text CLRSCR,NL
    !text "COLONY EDITOR USAGE:",NL,NL
    !TEXT "  CURSOR  - SELECT CELL",NL
    !TEXT "  SPACE   - TOGGLE CELL",NL
    !TEXT "  'C'     - CLEAR ALL CELLS",NL
    !TEXT "  'R'     - CREATE RANDOM COLONY",NL
    !TEXT "  'Q'     - EXIT EDITOR/START SIMULATION",NL,NL
    !text "AT ANY TIME DURING SIMULATION, PRESS:",NL,NL
    !text "  'Q'     - TO EXIT",NL
    !text "  'P'     - TO PAUSE",NL
    !TEXT "  'E'     - TO EDIT",NL
    !TEXT "  'M'     - TO RETURN TO START MENU",NL
    !text "  ANY KEY - TO RESUME FROM PAUSE",NL,NL
    !text "AT ANY TIME, PRESS 'H' TO SHOW THIS HELP",NL,NL
    !text "NOTE: SIMULATOR STARTS IN PAUSED MODE.",NL,0

; farewell message text, cbm ascii
goodbye

    !text "GOOD BYE! PLAY AGAIN SOON!",NL,0
    
; framed confirmation prompt text: "are you sure (y/n)?", screen codes
scrareyousure
    !byte $20,$20,$20,$20,$20,$20,$20,$20,$20,$70,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$6e,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20
    !byte $20,$20,$20,$20,$20,$20,$20,$20,$20,$42,$01,$12,$05,$20,$19,$0f,$15,$20,$13,$15,$12,$05,$20,$28,$19,$2f,$0e,$29,$3f,$42,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20
    !byte $20,$20,$20,$20,$20,$20,$20,$20,$20,$6d,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$7d,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20
    !byte 0

; framed confirmation prompt text: "save colony (y/n)?", screen codes 
scrsavecolony    
    !byte $20,$20,$20,$20,$20,$20,$20,$20,$20,$70,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$6e,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20
    !byte $20,$20,$20,$20,$20,$20,$20,$20,$20,$42,$13,$01,$16,$05,$20,$03,$0f,$0c,$0f,$0e,$19,$20,$28,$19,$2f,$0e,$29,$3f,$42,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20
    !byte $20,$20,$20,$20,$20,$20,$20,$20,$20,$6d,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$7d,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20
    !byte 0

; framed file name prompt text: "enter file name:             ", screen codes
scrfnameprompt

    !byte $20,$20,$20,$20,$70,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$6e,$20,$20,$20,$20,$20
    !byte $20,$20,$20,$20,$42,$05,$0e,$14,$05,$12,$20,$06,$09,$0c,$05,$20,$0e,$01,$0d,$05,$3a,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$20,$42,$20,$20,$20,$20,$20
    !byte $20,$20,$20,$20,$6d,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$40,$7d,$20,$20,$20,$20,$20
    !byte 0

; file name prompt text, cbm ascii, for loading from file
fnameprompt2    

    !text NL,NL,"ENTER FILE NAME:",0
    
; screen menu text, cbm ascii
menu1

    !text CLRSCR,CRSDOWN,CRSDOWN,CRSDOWN,CRSDOWN
    !text "     SELECT TEMPLATE COLONY:",NL,NL
    !text "      '1' - SHOOTER",NL
    !text "      '2' - OSCILLATOR 144",NL
    !text "      '3' - EMPTY",NL    
    !text "      '4' - RANDOM",NL,NL
    !text "              OR",NL,NL
    !text "      'L' - LOAD FROM FILE",NL
    !text "      'Q' - QUIT",0
    
; the not-implemented message    
;sorry1

;    !text NL,NL
;    !text "SORRY, THIS OPTION IS NOT IMPLEMENTED.",0
    
; colony file name extension    
fnamext

    !text ".COL",0
    
; variables, buffers

xind  !byte 0             ; row
yind  !byte 0             ; column
yold  !byte 0             ; column (previous)
nbct  !byte 0             ; neighbor cells counter
smct  !byte 0,0           ; current cell address
gpct  !byte 0,0           ; generic purpose counter
keyp  !byte 0             ; buffer for a pressed key code
gpbyt !byte 0             ; generic purpose byte buffer
fnamb !byte 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 ; file name buffer, 16-char + NULL
fnlen !byte 0             ; dedicated file name length register
; NOTE: keep this label at the end, it is a colony buffer (1000 bytes).
colb  !byte 0             ; colony buffer, 1000 bytes